LIBRARY "mtipli.dll"
EXPORTS
vpi_chk_error
vpi_compare_objects
vpi_control
vpi_flush
vpi_fopen
vpi_free_object
vpi_get
vpi_get_delays
vpi_get_file
vpi_get_str
vpi_get_systf_info
vpi_get_time
vpi_get_userdata
vpi_get_value
vpi_get_vlog_info
vpi_handle
vpi_handle_by_index
vpi_handle_by_name
vpi_iterate
vpi_mcd_close
vpi_mcd_flush
vpi_mcd_name
vpi_mcd_open
vpi_mcd_printf
vpi_mcd_vprintf
vpi_printf
vpi_put_delays
vpi_put_userdata
vpi_put_value
vpi_register_cb
vpi_register_systf
vpi_remove_cb
vpi_scan
vpi_sim_control
vpi_sim_vcontrol
vpi_vprintf
vpip_count_drivers
vpip_format_strength
vpip_make_systf_system_defined
vpip_mcd_rawwrite
vpip_set_return_value
acc_fetch_fullname
acc_fetch_fulltype
acc_fetch_name
acc_fetch_type
acc_fetch_type_str
mti_AddLoadDoneCB
mti_AddQuitCB
mti_Break
mti_CreateProcess
mti_CreateProcessWithPriority
mti_Delta
mti_Desensitize
mti_FindRegion
mti_FindSignal
mti_FindVar
mti_FirstLowerRegion
mti_FirstLowerRegion
mti_FirstSignal
mti_FirstVarByRegion
mti_ForceSignal
mti_GetArrayElementType
mti_GetArraySignalValue
mti_GetArrayVarValue
mti_GetEnumValues
mti_GetNumRecordElements
mti_GetPrimaryName
mti_GetProductVersion
mti_GetRegionFullName
mti_GetRegionName
mti_GetRegionSourceName
mti_GetResolutionLimit
mti_GetSignalName
mti_GetSignalNameIndirect
mti_GetSignalSubelements
mti_GetSignalType
mti_GetSignalValue
mti_GetSignalValueIndirect
mti_GetTopRegion
mti_GetTypeKind
mti_GetVarKind
mti_GetVarName
mti_GetVarSubelements
mti_GetVarType
mti_GetVarValue
mti_GetVarValueIndirect
mti_NextRegion
mti_NextSignal
mti_NextVar
mti_Now
mti_NowUpper
mti_Quit
mti_ReleaseSignal
mti_RemoveLoadDoneCB
mti_RemoveQuitCB
mti_ScheduleWakeup
mti_ScheduleWakeup64
mti_Sensitize
mti_SetSignalValue
mti_SetVarValue
mti_TickLeft
mti_TickLength
mti_TickRight
mti_VsimFree
mti_Interp
mti_Cmd
Tcl_GetStringResult
Tcl_ResetResult
Tcl_GetObjResult
Tcl_ResetResult
Tcl_ListObjGetElements
Tcl_GetStringResult
TclFreeObj
Tcl_ResetResult
Tcl_ResetResult
Tcl_GetString
TclFreeObj
vhpi_check_error
vhpi_control
vhpi_disable_cb
vhpi_enable_cb
vhpi_get
vhpi_get_phys
vhpi_get_str
vhpi_get_time
vhpi_get_value
vhpi_handle
vhpi_handle_by_index
vhpi_handle_by_name
vhpi_iterator
vhpi_put_value
vhpi_register_cb
vhpi_release_handle
vhpi_remove_cb
vhpi_scan
